<!doctype html>
<html lang="en">
    <head>
        <meta charset="UTF-8" />
        <meta name="viewport" content="width=device-width, initial-scale=1.0" />
        <meta http-equiv="X-UA-Compatible" content="ie=edge" />
        <title>EFMCalculator Results</title>
        <link rel="icon" type="image/x-icon" href="assets/favicon.svg" />
        <style></style>
        <script src="https://cdn.bokeh.org/bokeh/release/bokeh-{{bokehversion}}.min.js"></script>
        <script src="https://cdn.bokeh.org/bokeh/release/bokeh-widgets-{{bokehversion}}.min.js"></script>
        <script src="https://cdn.bokeh.org/bokeh/release/bokeh-tables-{{bokehversion}}.min.js"></script>
        <script src="https://cdn.bokeh.org/bokeh/release/bokeh-gl-{{bokehversion}}.min.js"></script>
        <script src="https://cdn.bokeh.org/bokeh/release/bokeh-mathjax-{{bokehversion}}.min.js"></script>
        <link rel="stylesheet" href="assets/stylesheet.css" />
    </head>
    <body>
        <div id="page-container">
            <nav>
                <div class="logo_wrapper">
                    <img
                        src="assets/cryptkeeper_long.svg"
                        alt="EFMCalculator Logo"
                        height="50px"
                    />
                </div>
            </nav>

            <main>
                <div class="bokehgraph">{{bokehdiv}}</div>
            </main>

            <footer></footer>
        </div>
        {{bokehscript}}
    </body>
</html>
