html,
body {
    margin: 0;
    padding: 0;
    position: relative;
    min-height: 100vh;
}

#content-wrap {
    padding-bottom: 2.5rem;
}

nav {
    background-color: #b9b9b9;
    border-bottom: 1px solid #e5e5e5;
    width: 100%;
    padding: 10px;
    height: 60px;
}

main {
    padding-left: 5%;
    padding-right: 5%;
    padding-top: 10px;
    padding-bottom: 2.5rem;
}

footer {
    position: absolute;
    bottom: 0;
    background-color: #b9b9b9;
    border-bottom: 1px solid #e5e5e5;
    width: 100%;
    height: 2.5rem;
}

div.bokehgraph {
    width: 90%;
    margin-left: auto;
    margin-right: auto;
    text-align: center;
}

.bk-root .bk {
    margin: 0 auto !important;
}
